"""Deconfounded ordinary least-squares based-estimators for causal inference."""
from time import perf_counter

# pylint:disable-msg=invalid-name
import numpy as np
from whynot.framework import InferenceResult

import whynot_estimators


class Deconfounder(whynot_estimators.Estimator):
    """Run deconfounded ols to estimate causal effect of treatment on the outcome y.

    Models the outcome variable Y as:
        Y = b_0 + b_1 * treatment + b_C^T * control.

    For more information on the deconfounder, please refer to:
        Wang, Yixin, and David M. Blei. "The blessings of multiple causes." arXiv
        preprint arXiv:1805.06826 (2018).

    """

    @property
    def name(self):
        """Estimator name."""
        return "deconfounder"

    def import_estimator(self):
        """Verify tensorflow dependencies are installed."""
        # pylint:disable-msg=unused-import
        import statsmodels.api
        import tensorflow
        from tensorflow_probability import edward2

    # pylint:disable-msg=too-many-locals,too-many-statements
    def estimate_treatment_effect(self, covariates, treatment, outcome):
        """Run deconfounded ols to estimate causal effect of treatment on the outcome y.

        Models the outcome variable Y as:
            Y = b_0 + b_1 * treatment + b_C^T * control.

        For more information on the deconfounder, please refer to:
            Wang, Yixin, and David M. Blei. "The blessings of multiple causes." arXiv
            preprint arXiv:1805.06826 (2018).

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for
                each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        import statsmodels.api as sm
        import tensorflow as tf
        from tensorflow_probability import edward2 as ed

        features = np.copy(covariates)
        treatment = treatment.reshape(-1, 1)
        features = np.concatenate([treatment, features], axis=1)

        def ppca_model(data_dim, latent_dim, num_datapoints, stddv_datapoints):
            """Return probabilistic pca model."""
            # pylint:disable-msg=no-member
            w = ed.Normal(
                loc=tf.zeros([latent_dim, data_dim]),
                scale=tf.ones([latent_dim, data_dim]),
                name="w",
            )  # parameter
            z = ed.Normal(
                loc=tf.zeros([num_datapoints, latent_dim]),
                scale=tf.ones([num_datapoints, latent_dim]),
                name="z",
            )  # local latent variable / substitute confounder
            x = ed.Normal(
                loc=tf.multiply(tf.matmul(z, w), 1),
                scale=stddv_datapoints * tf.ones([num_datapoints, data_dim]),
                name="x",
            )  # (modeled) data
            return x, (w, z)

        def variational_model(qw_mean, qw_stddv, qz_mean, qz_stddv):
            """Return variational model."""
            # pylint:disable-msg=no-member
            qw = ed.Normal(loc=qw_mean, scale=qw_stddv, name="qw")
            qz = ed.Normal(loc=qz_mean, scale=qz_stddv, name="qz")
            return qw, qz

        start_time = perf_counter()

        # do probabilistic PCA
        # pylint:disable-msg=unpacking-non-sequence
        log_joint = ed.make_log_joint_fn(ppca_model)
        latent_dim = 2
        stddv_datapoints = 0.1
        num_datapoints, data_dim = features.shape

        model = ppca_model(
            data_dim=data_dim,
            latent_dim=latent_dim,
            num_datapoints=num_datapoints,
            stddv_datapoints=stddv_datapoints,
        )

        def target(w, z):
            """Unnormalized target density as a function of the parameters."""
            return log_joint(
                data_dim=data_dim,
                latent_dim=latent_dim,
                num_datapoints=num_datapoints,
                stddv_datapoints=stddv_datapoints,
                w=w,
                z=z,
                x=features,
            )

        log_q = ed.make_log_joint_fn(variational_model)

        def target_q(qw, qz):
            return log_q(
                qw_mean=qw_mean,
                qw_stddv=qw_stddv,
                qz_mean=qz_mean,
                qz_stddv=qz_stddv,
                qw=qw,
                qz=qz,
            )

        qw_mean = tf.Variable(np.ones([latent_dim, data_dim]), dtype=tf.float32)
        qz_mean = tf.Variable(np.ones([num_datapoints, latent_dim]), dtype=tf.float32)
        qw_stddv = tf.nn.softplus(
            tf.Variable(-4 * np.ones([latent_dim, data_dim]), dtype=tf.float32)
        )
        qz_stddv = tf.nn.softplus(
            tf.Variable(-4 * np.ones([num_datapoints, latent_dim]), dtype=tf.float32)
        )

        qw, qz = variational_model(
            qw_mean=qw_mean, qw_stddv=qw_stddv, qz_mean=qz_mean, qz_stddv=qz_stddv
        )

        energy = target(qw, qz)
        entropy = -target_q(qw, qz)
        elbo = energy + entropy

        optimizer = tf.train.AdamOptimizer(learning_rate=0.05)
        train = optimizer.minimize(-elbo)

        init = tf.global_variables_initializer()

        t = []
        num_epochs = 500
        with tf.Session() as sess:
            sess.run(init)

            for i in range(num_epochs):
                sess.run(train)
                if i % 5 == 0:
                    t.append(sess.run([elbo]))

                z_mean_inferred = sess.run(qz_mean)

        # approximate the (random variable) substitute confounders with their inferred mean.
        Z_hat = z_mean_inferred
        # augment the regressors to be both the covariates and the substitute confounder Z
        features_aug = np.column_stack([features, Z_hat])

        # do standard OLS on augmented features
        features_aug = sm.add_constant(features_aug, prepend=True, has_constant="add")

        model = sm.OLS(outcome, features_aug)
        results = model.fit()

        stop_time = perf_counter()

        # Treatment is the second variable (first is the constant offset)
        ate = results.params[1]
        stderr = results.bse[1]
        conf_int = (ate - 1.96 * stderr, ate + 1.96 * stderr)
        return InferenceResult(
            ate=ate,
            stderr=stderr,
            ci=conf_int,
            individual_effects=None,
            elapsed_time=stop_time - start_time,
        )


DECONFOUNDER = Deconfounder()



"""Module level installation script."""
import os
import subprocess
import sys

import click

# Map of estimator name to python, CRAN, and R dependencies on github.
ESTIMATORS = {
    "causal_bart": {
        "python": ["rpy2"],
        "cran": ["dbarts"],
        "github_r": ["vdorie/bartCause"],
    },
    "causal_forest": {"python": ["rpy2"], "cran": ["grf"],},
    "deconfounder": {"python": ["tensorflow==1.*", "tensorflow_probability<=0.8"],},
    "doubleml": {"python": ["econml"],},
    "ip_weighting": {"python": ["rpy2"], "cran": ["WeightIt", "survey"],},
    "matching": {"python": ["rpy2"], "cran": ["Matching"],},
    "rlearner": {"python": ["causalml"],},
    "slearner": {"python": ["causalml"],},
    "tmle": {"python": ["rpy2"], "cran": ["tmle", "dbarts", "gam"]},
}


@click.group()
def main():
    """Entry point for CLI."""


def install_python(package):
    """Install a python package using PIP."""
    if package == "rpy2" and not r_build_found():
        print("Cannot install rpy2!")
        raise ValueError("Unable to find R installation. Please install R and retry.")

    cmd = [sys.executable, "-m", "pip", "install", "--upgrade", package]
    return subprocess.check_call(cmd, env=os.environ.copy())


def r_build_found():
    """Return whether or not an R build is accessible.

    Works by trying to find R's home directory (aka R_HOME).
    First it check for an environment variable R_HOME, and
    and if none is found it tries to get it from an R executable
    in the PATH.

    See: https://github.com/rpy2/rpy2/blob/master/rpy2/situation.py.
    """
    if os.environ.get("R_HOME"):
        return True
    try:
        _ = subprocess.check_output(("R", "RHOME"), universal_newlines=True)
    # pylint:disable-msg=broad-except
    except Exception:  # FileNotFoundError, WindowsError, etc
        return False
    return True


def install_r_package(package, location="CRAN"):
    """Install R packages from CRAN.

    Parameters
    ----------
        package: str
            Name of the package to install.
        location: str
            Either CRAN or GITHUB

    """
    if location not in ["CRAN", "GITHUB"]:
        raise ValueError(f"location {location} must be either CRAN or GITHUB.")

    # Ensure R and rpy2 are accessible.
    if not r_build_found():
        msg = (
            "Unable to find R installation. Please install R and retry.\n"
            "Installation instructions available at: "
            "https://github.com/zykls/whynot_estimators"
        )
        raise ValueError(msg)
    try:
        import rpy2.robjects.packages as rpackages
    except ImportError:
        raise ValueError(
            "Unable to import rpy2. rpy2 is required to install R packages."
        )

    if rpackages.isinstalled(package):
        return

    # pylint:disable-msg=no-member
    utils = rpackages.importr("utils")
    utils.chooseCRANmirror(ind=1)
    if location == "CRAN":
        utils.install_packages(package)
    else:
        if not rpackages.isinstalled("remotes"):
            utils.install_packages("remotes")
        remotes = rpackages.importr("remotes")
        remotes.install_github(package)


@main.command("show_all")
def show_all():
    """Show all of the estimators available in whynot_estimators."""
    print("Name \t\t\t\t R Required?")
    for estimator, dependencies in ESTIMATORS.items():
        requires_r = False
        if "cran" in dependencies and dependencies["cran"]:
            requires_r = True
        elif "github_r" in dependencies and dependencies["github_r"]:
            requires_r = True

        print(f"{estimator} \t\t\t {requires_r}")


@main.command("install")
@click.argument("estimator", type=str)
def install(estimator):
    """Install an estimator in whynot."""
    if estimator not in ESTIMATORS and estimator != "all":
        all_estimators = "\n\t".join(ESTIMATORS.keys())
        msg = f"Estimator {estimator} not currently supported.\n Available:\n\t{all_estimators}"
        raise ValueError(msg)

    def run_install(estimator_name):
        dependencies = ESTIMATORS[estimator_name]
        if "python" in dependencies:
            for package in dependencies["python"]:
                install_python(package)
        if "cran" in dependencies:
            for package in dependencies["cran"]:
                install_r_package(package, location="CRAN")
        if "github_r" in dependencies:
            for package in dependencies["github_r"]:
                install_r_package(package, location="GITHUB")

    if estimator == "all":
        for estimator_name in ESTIMATORS:
            try:
                run_install(estimator_name)
            # pylint:disable-msg=broad-except
            except Exception:
                print("=" * 80)
                print(f"Failed to install estimator {estimator_name}!")
                print("=" * 80)
                continue
    else:
        run_install(estimator)


if __name__ == "__main__":
    main()



"""Estimate treatment effects with R-learning."""
from time import perf_counter

from whynot.framework import InferenceResult

import whynot_estimators


class RLearnerMLP(whynot_estimators.Estimator):
    """Estimate treatment effects with an R-learner and MLP regressor."""

    @property
    def name(self):
        """Estimator name."""
        return "rlearner_mlp"

    def import_estimator(self):
        """Import the causalml package."""
        # pylint:disable-msg=unused-import
        import causalml

    def estimate_treatment_effect(
        self,
        covariates,
        treatment,
        outcome,
        hidden_layer_sizes=(10, 10),
        learning_rate_init=0.1,
        early_stopping=True,
    ):
        """Estimate treatment effect with an R-learner and a MLP regressor.

        Wrapper around causalml.inference.meta.MLPTRegressor.

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Binary array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        from causalml.inference.meta import MLPTRegressor

        start_time = perf_counter()
        nn_regressor = MLPTRegressor(
            hidden_layer_sizes=hidden_layer_sizes,
            learning_rate_init=learning_rate_init,
            early_stopping=early_stopping,
        )
        ate, lower_bound, upper_bound = nn_regressor.estimate_ate(
            covariates, treatment, outcome
        )
        stop_time = perf_counter()

        return InferenceResult(
            ate=ate[0],
            stderr=None,
            ci=(lower_bound[0], upper_bound[0]),
            individual_effects=None,
            elapsed_time=stop_time - start_time,
        )


RLEARNER_MLP = RLearnerMLP()



"""Causal estimators based on Bayesian Additive Regression Trees."""
from time import perf_counter

from whynot.framework import InferenceResult

import whynot_estimators
import whynot_estimators.utils as utils


class CausalBart(whynot_estimators.Estimator):
    """Estimate average and individual treatment effects using causal BART.

    The estimator reprises on the `bartCause` package. For more details see:

        Hill, J. L. (2011) Bayesian Nonparametric Modeling for Causal Inference. Journal
        of Computational and Graphical Statistics 20(1), 217–240. Taylor & Francis.
        https://doi.org/10.1198/jcgs.2010.08162.

    """

    @property
    def name(self):
        """Estimator name."""
        return "causal_bart"

    def import_estimator(self):
        """Import the BART package."""
        import rpy2.robjects.packages as rpackages
        from rpy2.robjects import numpy2ri

        numpy2ri.activate()
        if not rpackages.isinstalled("bartCause"):
            raise ImportError(f"Package {self.name} is not installed!")
        # pylint:disable-msg=attribute-defined-outside-init
        self.bartcause = rpackages.importr("bartCause")

    def estimate_treatment_effect(self, covariates, treatment, outcome):
        """Estimate average and individual treatment effects using causal BART.

        The estimator reprises on the `bartCause` package. For more details see:

            Hill, J. L. (2011) Bayesian Nonparametric Modeling for Causal Inference. Journal
            of Computational and Graphical Statistics 20(1), 217–240. Taylor & Francis.
            https://doi.org/10.1198/jcgs.2010.08162.

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for
                each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        start_time = perf_counter()
        # pylint:disable-msg=no-member
        model = self.bartcause.bartc(outcome, treatment, covariates, verbose=False)
        results = self.bartcause.summary_bartcFit(model)
        stop_time = perf_counter()

        estimates = utils.extract(results, "estimates")
        ate = utils.extract(estimates, "estimate")[0]
        stderr = utils.extract(estimates, "sd")[0]
        ci_lower = utils.extract(estimates, "ci.lower")[0]
        ci_upper = utils.extract(estimates, "ci.upper")[0]

        return InferenceResult(
            ate=ate,
            stderr=stderr,
            ci=[ci_lower, ci_upper],
            individual_effects=None,
            elapsed_time=stop_time - start_time,
        )


CAUSAL_BART = CausalBart()



"""Causal estimators based on Targeted Maximum Likelihood Estimation (TMLE)."""
from time import perf_counter

import numpy as np
from whynot.framework import InferenceResult

import whynot_estimators
import whynot_estimators.utils as utils


class Tmle(whynot_estimators.Estimator):
    """Estimates average and individual treatment effects using TMLE.

    This estimator uses the `tmle` package to perform targeted maximum likelihood
    estimation. For more details, see:

        https://cran.r-project.org/web/packages/tmle/tmle.pdf

    For a brief introduction to these techniques, see:

        Gruber, Susan and van der Laan, Mark J., "Targeted Maximum Likelihood
        Estimation: A Gentle Introduction" (August 2009). U.C. Berkeley Division of
        Biostatistics Working Paper Series. Working Paper 252.
        https://biostats.bepress.com/ucbbiostat/paper252

    """

    @property
    def name(self):
        """Estimator name."""
        return "tmle"

    def import_estimator(self):
        """Import the estimator and its dependencies."""
        import rpy2.robjects.packages as rpackages
        from rpy2.robjects import numpy2ri

        numpy2ri.activate()
        if not rpackages.isinstalled("tmle"):
            raise ImportError(f"Package {self.name} is not installed!")
        self.tmle = rpackages.importr("tmle")

    def estimate_treatment_effect(self, covariates, treatment, outcome):
        """Estimates average and individual treatment effects using TMLE.

        This estimator uses the `tmle` package to perform targeted maximum likelihood
        estimation. For more details, see:
            https://cran.r-project.org/web/packages/tmle/tmle.pdf

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Array of shape [num_samples]  indicating treatment status for each sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        # Ensure arrays will be conformable.
        treatment = treatment.reshape(-1, 1)
        outcome = outcome.reshape(-1, 1)

        start_time = perf_counter()
        # pylint:disable-msg=no-member
        model = self.tmle.tmle(outcome, treatment, covariates)
        stop_time = perf_counter()

        # From the TMLE documentation
        # https://cran.r-project.org/web/packages/tmle/tmle.pdf
        estimates = utils.extract(model, "estimates")
        ate_estimate = utils.extract(estimates, "ATE")
        ate = utils.extract(ate_estimate, "psi")
        conf_int = utils.extract(ate_estimate, "CI")
        var = utils.extract(ate_estimate, "var.psi")

        # Unpack
        ate = float(ate[0])
        var = float(var[0])
        conf_int = (float(conf_int[0]), float(conf_int[1]))

        return InferenceResult(
            ate=ate,
            stderr=np.sqrt(var),
            ci=conf_int,
            individual_effects=None,
            elapsed_time=stop_time - start_time,
        )


TMLE = Tmle()



"""Causal estimators based on random forests."""
from time import perf_counter

from whynot.framework import InferenceResult

import whynot_estimators


class CausalForest(whynot_estimators.Estimator):
    """Estimate average and individual treatment effects using causal forest.

    The causal forest implementation uses the `grf` package. For more details see:

        Susan Athey, Julie Tibshirani and Stefan Wager. Generalized Random Forests,
        Annals of Statistics 47.2 (2019): 1148-1178.

    """

    @property
    def name(self):
        """Estimator name."""
        return "causal_forest"

    def import_estimator(self):
        """Import the grf package."""
        # pylint:disable-msg=import-outside-toplevel
        import rpy2.robjects.packages as rpackages
        from rpy2.robjects import numpy2ri

        numpy2ri.activate()
        if not rpackages.isinstalled("grf"):
            raise ImportError(f"Package {self.name} is not installed")
        # pylint:disable-msg=attribute-defined-outside-init
        self.grf = rpackages.importr("grf")

    def estimate_treatment_effect(self, covariates, treatment, outcome):
        """Estimate average and individual treatment effects using causal forest.

        The causal forest implementation uses the `grf` package. For more details see:

            Susan Athey, Julie Tibshirani and Stefan Wager. Generalized Random Forests,
            Annals of Statistics 47.2 (2019): 1148-1178.

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for
                each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        # Convert to matrices
        treatment = treatment.reshape(-1, 1)
        outcome = outcome.reshape(-1, 1)

        start_time = perf_counter()
        # rpy2 automatically handles the conversion of numpy
        # arrays into r objects. The returned forest object is an rpy2 object!
        # pylint:disable-msg=no-member
        forest = self.grf.causal_forest(
            covariates, outcome, treatment, tune_parameters="all"
        )

        # Estimate the conditional average treatment effect. rpy2 semantics
        # so target.sample in the R api becomes target_sample
        # pylint:disable-msg=no-member
        result = self.grf.average_treatment_effect(forest, target_sample="all")
        cate, stderr = result[0], result[1]

        predictions = self.grf.predict_causal_forest(forest, covariates)

        stop_time = perf_counter()

        individual_effects = [item for sublist in predictions for item in sublist]

        conf_int = [cate - 1.96 * stderr, cate + 1.96 * stderr]
        return InferenceResult(
            ate=cate,
            stderr=stderr,
            ci=conf_int,
            individual_effects=individual_effects,
            elapsed_time=stop_time - start_time,
        )


CAUSAL_FOREST = CausalForest()



"""Algorithms for causal inference."""

__version__ = "0.10.13"
from whynot_estimators.estimator import Estimator
from whynot_estimators.bart import CAUSAL_BART as causal_bart
from whynot_estimators.causal_forest import CAUSAL_FOREST as causal_forest
from whynot_estimators.deconfounder import DECONFOUNDER as deconfounder
from whynot_estimators.doubleml import DOUBLEML as doubleml
from whynot_estimators.ip_weighting import IP_WEIGHTING as ip_weighting
from whynot_estimators.matching import MATCHING as matching
from whynot_estimators.rlearner_mlp import RLEARNER_MLP as rlearner_mlp
from whynot_estimators.rlearner_xbg import RLEARNER_XBG as rlearner_xbg
from whynot_estimators.slearner_linear import SLEARNER_LINEAR as slearner_linear
from whynot_estimators.tmle import TMLE as tmle

ESTIMATORS = [
    causal_bart,
    causal_forest,
    deconfounder,
    doubleml,
    ip_weighting,
    matching,
    rlearner_mlp,
    rlearner_xbg,
    slearner_linear,
    tmle,
]


def get_installed():
    """Return all of the currently installed estimators."""
    return [estimator for estimator in ESTIMATORS if estimator.is_installed()]



"""Double machine learning methods for causal inference."""
from time import perf_counter

import numpy as np
from sklearn.linear_model import LogisticRegressionCV
from sklearn.ensemble import RandomForestRegressor
from whynot.framework import InferenceResult

import whynot_estimators


class DoubleML(whynot_estimators.Estimator):
    """Use double machine learning method to estimate treatment effects."""

    @property
    def name(self):
        """Estimator name."""
        return "double_ml"

    def import_estimator(self):
        """Import the econml package."""
        # pylint:disable-msg=unused-import
        import econml

    def estimate_treatment_effect(self, covariates, treatment, outcome):
        """Run a double machine learning method to estimate treatment effect.

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for
                each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        from econml.dml import DMLCateEstimator

        start_time = perf_counter()

        # TODO: There are many choices for these estimators we should expose.
        estimator = DMLCateEstimator(
            model_y=RandomForestRegressor(n_estimators=10),
            model_t=LogisticRegressionCV(cv=5, max_iter=200),
            discrete_treatment=True,
        )

        # X is features/axes of heterogeneity, W is controls. We assume
        # assume that we care about heterogeneity for every covariate.
        estimator.fit(outcome, treatment, X=covariates, W=None)

        # Compute treatment effects on the sample
        treatment_effects = estimator.effect(T0=0, T1=1, X=covariates)
        ate = np.mean(treatment_effects)

        stop_time = perf_counter()

        # TODO: econml will soon support bootstrap uncertainty estimation.
        return InferenceResult(
            ate=ate,
            stderr=None,
            ci=None,
            individual_effects=treatment_effects,
            elapsed_time=stop_time - start_time,
        )


DOUBLEML = DoubleML()



"""Causal estimators based on IP weighting techniques."""
from time import perf_counter

import pandas as pd
from whynot.framework import InferenceResult

import whynot_estimators
import whynot_estimators.utils as utils


class IPWeighting(whynot_estimators.Estimator):
    """Estimate average treatment effects via inverse propensity scores.

    Uses WeightIt to get propensity scores:
        https://cran.r-project.org/web/packages/WeightIt/WeightIt.pdf

    Uses Survey to fit a generalized linear model:
        https://cran.r-project.org/web/packages/survey/survey.pdf

    """

    @property
    def name(self):
        """Estimator name."""
        return "ip_weighting"

    def import_estimator(self):
        """Import WeightIt and survey packages."""
        import rpy2.robjects.packages as rpackages
        from rpy2.robjects import numpy2ri, pandas2ri

        numpy2ri.activate()
        pandas2ri.activate()
        if not rpackages.isinstalled("WeightIt") or not rpackages.isinstalled("survey"):
            raise ImportError(f"Estimator {self.name} is not installed!")
        self.weightit = rpackages.importr("WeightIt")
        self.survey = rpackages.importr("survey")

    def estimate_treatment_effect(self, covariates, treatment, outcome):
        """Estimate average treatment effects via inverse propensity scores.

        Uses WeightIt to get propensity scores:
            https://cran.r-project.org/web/packages/WeightIt/WeightIt.pdf

        Uses Survey to fit a generalized linear model:
            https://cran.r-project.org/web/packages/survey/survey.pdf

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for
                each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        from rpy2.robjects import Formula

        # Convert data into a dataframe
        column_names = ["x_{}".format(idx) for idx in range(covariates.shape[1])]
        data = pd.DataFrame(covariates, columns=column_names)
        data["treatment"] = treatment
        data["outcome"] = outcome

        propensity_formula = Formula(f"treatment ~ {' + '.join(column_names)}")

        start_time = perf_counter()

        # Note: The field ["ps"] contains the actual propensity scores we could use
        # elsewhere if needed. Stabilize=True not strictly needed for reasonable
        # performance.
        # pylint:disable-msg=no-member
        weight_obj = self.weightit.weightit(
            propensity_formula, data=data, method="ps", estimand="ATE", stabilize=True
        )
        weights = utils.extract(weight_obj, "weights")

        # Use survey to fit weighted glm and get robust confidence intervals.
        design = self.survey.svydesign(Formula("~ 1"), weights=weights, data=data)
        fit = self.survey.svyglm(Formula("outcome ~ treatment"), design=design)

        # coefs[0] is bias term, coefs[1] is the treatment
        ate = self.survey.coef_svyglm(fit)[1]

        # 95% confidence interval for the treatment
        # cis[0] is for the bias term, cis[1] is for treatment.
        robust_ci = self.survey.confint_svyglm(fit)[1]
        robust_ci = (robust_ci[0], robust_ci[1])

        stop_time = perf_counter()

        return InferenceResult(
            ate=ate,
            stderr=None,
            ci=robust_ci,
            individual_effects=None,
            elapsed_time=stop_time - start_time,
        )


IP_WEIGHTING = IPWeighting()



"""Estimate treatment effects with R-learning and XBG."""
from time import perf_counter

from whynot.framework import InferenceResult

import whynot_estimators


class RLearnerXBG(whynot_estimators.Estimator):
    """Estimate treatment effects with an R-learner and a decision tree regressor."""

    @property
    def name(self):
        """Estimator name."""
        return "rlearner_xbg"

    def import_estimator(self):
        """Import the causalml package."""
        # pylint:disable-msg=unused-import
        import causalml

    def estimate_treatment_effect(self, covariates, treatment, outcome):
        """Estimate treatment effect with R-learner and boosted decision trees.

        Wrapper around causalml.inference.meta.XGBTRegressor.

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Binary array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        from causalml.inference.meta import XGBTRegressor

        start_time = perf_counter()
        xbg_regressor = XGBTRegressor()
        ate, lower_bound, upper_bound = xbg_regressor.estimate_ate(
            covariates, treatment, outcome
        )
        stop_time = perf_counter()

        return InferenceResult(
            ate=ate[0],
            stderr=None,
            ci=(lower_bound[0], upper_bound[0]),
            individual_effects=None,
            elapsed_time=stop_time - start_time,
        )


RLEARNER_XBG = RLearnerXBG()



"""Causal estimators based on matching techniques."""
from time import perf_counter

from whynot.framework import InferenceResult

import whynot_estimators
import whynot_estimators.utils as utils


class Matching(whynot_estimators.Estimator):
    """Estimate average treatment effect via matching.

    The estimator uses the `Matching` package. For more details, see
        https://cran.r-project.org/web/packages/Matching/Matching.pdf.

    """

    @property
    def name(self):
        """Estimator name."""
        return "matching"

    def import_estimator(self):
        """Import the matching package."""
        import rpy2.robjects.packages as rpackages
        from rpy2.robjects import numpy2ri

        numpy2ri.activate()
        if not rpackages.isinstalled("Matching"):
            raise ImportError(f"'Matching' package not installed!")
        self.matching = rpackages.importr("Matching")

    def estimate_treatment_effect(
        self,
        covariates,
        treatment,
        outcome,
        distance="mahalanobis",
        errors="abadie_imbens",
    ):
        """Estimate treatment effects using matching.

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for
                each sample.
            distance:
                One of "inverse_variance", "mahalanobis", or "multivariate".
            errors:
                One of "abadie-imbens" or the usual "neyman" standard errors.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")

        weight_matrix = None
        if distance == "inverse_variance":
            weight = 1
        elif distance == "mahalanobis":
            weight = 2
        elif distance == "multivariate":
            weight = 3
            # Use GenMatch to automatically find covariate balance
            # pylint:disable-msg=no-member
            genmatch_out = self.matching.GenMatch(
                treatment, covariates, estimand="ATE", verbose=False
            )
            weight_matrix = genmatch_out[2]
        else:
            raise ValueError(f"{distance} not a valid option for matching estimator.")

        start_time = perf_counter()
        # pylint: disable-msg=no-member
        if weight_matrix is None:
            results = self.matching.Match(
                outcome, treatment, covariates, estimand="ATE", Weight=weight
            )
        else:
            results = self.matching.Match(
                outcome,
                treatment,
                covariates,
                estimand="ATE",
                Weight=weight,
                Weight_matrix=weight_matrix,
            )

        stop_time = perf_counter()

        ate = float(utils.extract(results, "est")[0])
        if errors == "abadie_imbens":
            stderr = float(utils.extract(results, "se")[0])
        elif errors == "neyman":
            stderr = float(utils.extract(results, "se.standard")[0])
        else:
            raise ValueError(f"{errors} not a valid option.")

        conf_int = (ate - 1.96 * stderr, ate + 1.96 * stderr)
        return InferenceResult(
            ate=ate,
            individual_effects=None,
            stderr=stderr,
            ci=conf_int,
            elapsed_time=stop_time - start_time,
        )


MATCHING = Matching()



"""Build whynot_estimators package."""
from setuptools import setup, find_packages

setup(
    name="whynot_estimators",
    version="0.10.13",
    author="John Miller",
    author_email="miller_john@berkeley.edu",
    description="Companion package to whynot: A collection of causal estimators.",
    packages=find_packages(),
    install_requires=["cython", "numpy", "pandas", "sklearn", "whynot", "tzlocal",],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
    ],
    python_requires=">=3.6",
)



"""Tests for causal inference methods."""
import numpy as np

import whynot
import whynot_estimators


def generate_data(num_samples, num_features):
    """Generate a synthetic causal inference dataset."""
    covariates = np.random.randn(num_samples, num_features)
    treatment = (np.random.rand(num_samples) < 0.5).astype(np.int64)
    outcome = np.random.randn(num_samples)
    return covariates, treatment, outcome


def test_bart_basic():
    """Ensure BART algorithm runs."""
    covariates, treatment, outcome = generate_data(100, 10)
    if whynot_estimators.causal_bart.is_installed():
        _ = whynot_estimators.causal_bart.estimate_treatment_effect(
            covariates, treatment, outcome
        )


def test_causal_forest_basic():
    """Ensure causal forest algorithm runs."""
    covariates, treatment, outcome = generate_data(100, 10)
    _ = whynot_estimators.causal_forest.estimate_treatment_effect(
        covariates, treatment, outcome
    )


def test_deconfounder_basic():
    """Ensure deconfounders runs."""
    covariates, treatment, outcome = generate_data(100, 10)
    if whynot_estimators.deconfounder.is_installed():
        _ = whynot_estimators.deconfounder.estimate_treatment_effect(
            covariates, treatment, outcome
        )


def test_doubleml_basic():
    """Ensure doubleml runs."""
    covariates, treatment, outcome = generate_data(100, 10)
    if whynot_estimators.doubleml.is_installed():
        _ = whynot_estimators.doubleml.estimate_treatment_effect(
            covariates, treatment, outcome
        )


def test_ipweighting_basic():
    """Ensure ip weighting runs."""
    covariates, treatment, outcome = generate_data(100, 10)
    if whynot_estimators.ip_weighting.is_installed():
        _ = whynot_estimators.ip_weighting.estimate_treatment_effect(
            covariates, treatment, outcome
        )


def test_matching_basic():
    """Ensure matching runs."""
    covariates, treatment, outcome = generate_data(100, 10)
    if whynot_estimators.matching.is_installed():
        _ = whynot_estimators.matching.estimate_treatment_effect(
            covariates, treatment, outcome
        )


def test_rlearner_basic():
    """Ensure rlearner methods run."""
    covariates, treatment, outcome = generate_data(100, 10)
    if whynot_estimators.rlearner_mlp.is_installed():
        _ = whynot_estimators.rlearner_mlp.estimate_treatment_effect(
            covariates, treatment, outcome
        )
    if whynot_estimators.rlearner_xbg.is_installed():
        _ = whynot_estimators.rlearner_xbg.estimate_treatment_effect(
            covariates, treatment, outcome
        )


def test_slearner_basic():
    """Ensure slearner methods run."""
    covariates, treatment, outcome = generate_data(100, 10)
    if whynot_estimators.slearner_linear.is_installed():
        _ = whynot_estimators.slearner_linear.estimate_treatment_effect(
            covariates, treatment, outcome
        )


def test_tmle_basic():
    """Ensure tmle runs."""
    covariates, treatment, outcome = generate_data(100, 10)
    _ = whynot_estimators.tmle.estimate_treatment_effect(covariates, treatment, outcome)


def test_causal_suite():
    """Ensure casual suite picks up everything."""
    covariates, treatment, outcome = generate_data(100, 10)
    results = whynot.causal_suite(covariates, treatment, outcome)
    if whynot_estimators.causal_forest.is_installed():
        assert "causal_forest" in results
    if whynot_estimators.ip_weighting.is_installed():
        assert "ip_weighting" in results
    if whynot_estimators.matching.is_installed():
        assert "matching" in results
    if whynot_estimators.tmle.is_installed():
        assert "tmle" in results



"""Utitility tools for working with whynot_estimators."""


def extract(listvector, argname):
    """Retrieve argument argname from R listvector."""
    if argname not in listvector.names:
        raise ValueError(f"{argname} not found.")
    index = list(listvector.names).index(argname)
    return listvector[index]



"""Base class for whynot_estimators."""
from abc import abstractmethod, ABCMeta


class Estimator:
    """Generic estimator class."""

    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def name(self):
        """Name of the estimator."""
        return NotImplemented

    @abstractmethod
    def import_estimator(self):
        """Import the estimator."""
        return NotImplemented

    def is_installed(self):
        """Check if the estimator is installed."""
        try:
            self.import_estimator()
            return True
        except ImportError:
            return False

    @abstractmethod
    def estimate_treatment_effect(
        self, covariates, treatment, outcome, *args, **kwargs
    ):
        """Estimate average and individual treatment effects.

        Parameters
        ----------
            covariates: `np.ndarray`
                Array of shape [num_samples, num_features] of features
            treatment:  `np.ndarray`
                Array of shape [num_samples]  indicating treatment status for each
                sample.
            outcome:  `np.ndarray`
                Array of shape [num_samples] containing the observed outcome for
                each sample.

        Returns
        -------
            result: `whynot.framework.InferenceResult`
                InferenceResult object for this procedure

        """
        if not self.is_installed():
            raise ValueError(f"Estimator {self.name} is not installed!")
